def environment_variables():
    return {
        "GIT_AUTHOR_NAME": "AI Migrator",
        "GIT_AUTHOR_EMAIL": "dx-ai-tools--ai-migrator@block.xyz",
        "GIT_COMMITTER_NAME": "AI Migrator",
        "GIT_COMMITTER_EMAIL": "dx-ai-tools--ai-migrator@block.xyz",
    }
